import platform
import socket

from .logger import logger


def add_info(obj):
    try:
        obj.hostname = socket.gethostname()
    except Exception:
        logger.warning("Failed to get hostname", exc_info=1)
        obj.hostname = ""

    with socket.socket(socket.AF_INET, socket.SOCK_DGRAM) as s:
        try:
            # doesn't even have to be reachable
            s.connect(('10.255.255.255', 1))
            obj.ip = s.getsockname()[0]
        except Exception:
            logger.warning("Failed to get ip", exc_info=1)
            obj.ip = ""

    obj.machine_type = platform.machine()
    obj.network = platform.node()

    obj.os = platform.system()
    obj.os_release = platform.release()
    obj.os_version = platform.version()
    obj.os_string = platform.platform()



import sys
import time
import threading

import six
import os
import inspect
import site
import imp
import hashlib

from six.moves import builtins

from rook.logger import logger

from rook.config import ImportServiceConfig

from rook.serverless import on_lambda

from rook.exceptions import RookDependencyError, RookSourceFilePathSuggestion, RookDuplicatedModule

from rook.file_utils import FileUtils

from rook.processor.error import Error

_SUPPORTED_PY_FORMATS = ['.py', '.pyc']

_BLACKLISTED_PATHS = []

_OLD_IMPORT_FUNC = builtins.__import__


# only used in tests
def set_old_import_func(import_func):
    global _OLD_IMPORT_FUNC
    current = _OLD_IMPORT_FUNC
    _OLD_IMPORT_FUNC = import_func
    return current


class CountingImportLock(object):
    """
    When locked, only this thread can import (so all imports are guaranteed to have completed,
                                              and there are no partially initialized modules in sys.modules)
    The import lock is used here because:
    (1) it's reentrant, so we won't block user threads when they already hold it
    (2) it guarantees that if it's locked, no imports are partially complete
    (3) generally speaking, if you use more than 1 lock, there's the potential for deadlocks:
        locking (A) then (B) and another thread locking (B) then (A)
        This is an inherent risk in an import hook because the import lock is locked by Python itself,
        so using _any other lock_can  result in deadlocks
    """
    def __init__(self):
        self._count = 0

    def __enter__(self):
        imp.acquire_lock()
        self._count += 1

    def __exit__(self, *args, **kwargs):
        self._count -= 1
        imp.release_lock()

    def get_recursion_count(self):
        return self._count


class ImportService(object):

    NAME = "Import"

    def __init__(self, bdb_location_service):
        self._bdb_location_service = bdb_location_service

        self._modules = sys.modules.copy()
        self._path_cache = {}
        self._post_import_notifications = {}

        self._thread = None
        self._quit = False

        self.file_utils = FileUtils()

        external_paths = [sys.prefix]

        if hasattr(sys, 'base_prefix') and sys.base_prefix not in external_paths:
            external_paths += [sys.base_prefix]

        if hasattr(site, 'getsitepackages') and hasattr(site, 'getsitepackages') and hasattr(site, 'getusersitepackages'):
            external_paths += site.getsitepackages() + [site.getuserbase() + site.getusersitepackages()]

        self.external_paths = [os.path.normcase(os.path.realpath(external_path))
                               for external_path in external_paths]

        # pyenv messes up sys.path, so we ignore everything under it's folder
        for path in self.external_paths:
            pyenv_offset = path.find('.pyenv')
            if pyenv_offset != -1:
                pyenv_directory = path[:pyenv_offset + len('.pyenv') + 1]
                self.external_paths += [pyenv_directory]
                break

        # If we are (heuristicly) worried we missed anything, add "catch-alls"
        if len(self.external_paths) < 3:
            logger.debug("Missing getsitepackages, adding heuristics")
            self.external_paths += ['site-packages', 'site-python']

        logger.debug("External paths are " + str(self.external_paths))

        if on_lambda():
            _BLACKLISTED_PATHS.extend(['/var/task/flask', '/var/task/chalice'])

        import platform
        if platform.system() != 'Windows' and ImportServiceConfig.USE_IMPORT_HOOK:
            platform = platform.python_implementation().lower()

            def import_hook(*args, **kwargs):
                """
                Declared here to make it obvious that you can't replace the function by monkeypatching -
                the C extension will always reference the function set with SetImportHook.
                """
                __rookout__tracebackhide__ = True
                # Locking the global import lock here extends the section for which it is locked -
                # we don't want the import hook to be called from two threads simultaneously,
                # but the scope in which the import lock is held might only encompass the call to the original
                # __import__. The import lock is re-entrant, so it's OK for us to lock it and then for Python
                # to lock it again.
                with import_hook.import_lock:
                    result = _OLD_IMPORT_FUNC(*args, **kwargs)

                    # TODO - review if we can process the module being loaded regardless instead of evaluating the full list
                    try:
                        # if the recursion count is > 1,
                        # then this call to import_hook is a nested import, which means that the parent
                        # import has not finished executing its code yet, but it's already in sys.modules.
                        # if we evaluate now, we will try to place BPs in the module before it finished executing,
                        # potentially resulting in CodeNotFound.
                        if import_hook.import_lock.get_recursion_count() == 1:
                            self.evaluate_module_list()
                    except:  # lgtm[py/catch-base-exception]
                        pass

                    return result
            # this must be a reentrant lock - imports can cause other imports
            import_hook.import_lock = CountingImportLock()

            if platform == "cpython":
                try:
                    import native_extensions
                except Exception as e:
                    raise RookDependencyError(e)

                logger.debug('Enabling native import hook')
                native_extensions.SetImportHook(import_hook)
                # atomic swap
                builtins.__import__ = native_extensions.CallImportHookRemovingFrames
            elif platform == "pypy":
                import __pypy__
                logger.debug('Enabling pypy import hook')
                # atomic swap
                builtins.__import__ = __pypy__.hidden_applevel(import_hook)
            else:
                # assertion should never be reached, singleton.py checks platform support
                raise AssertionError("Unsupported platform")
        else:
            self._thread = threading.Thread(target=self._query_thread,
                                            name=ImportServiceConfig.THREAD_NAME)
            self._thread.daemon = True
            self._thread.start()

    def close(self):
        if _OLD_IMPORT_FUNC:
            builtins.__import__ = _OLD_IMPORT_FUNC

        if self._thread:
            self._quit = True

            # If threading was monkey patched by gevent waiting on thread will likely throw an exception
            try:
                from gevent.monkey import is_module_patched
                if is_module_patched("threading"):
                    time.sleep(ImportServiceConfig.SYS_MODULES_QUERY_INTERVAL)
            except Exception:  # Nothing we can do here but join as usual
                pass

            self._thread.join()

    def register_post_import_notification(self, location):
        filepath = location.filename
        filename = os.path.basename(filepath) if filepath else None

        # Register notification for future loads
        self._post_import_notifications[location.aug_id] = location

        # declaring filenames blacklist for avoiding setting multiple callbacks on the same aug
        modules_blacklist = []
        match_found = False
        file_hashes = {}
        with CountingImportLock():
            # Attempt to satisfy location using known modules
            for module_key, module_object in self._modules.items():
                # Get module details and check if it matches
                module_filename = self._get_module_path(module_object)

                # If module is not valid, ignore
                if not self._is_valid_module(module_object, module_filename):
                    continue

                if filename == os.path.basename(module_object.__file__):
                    file_content = self.file_utils.get_safe_file_contents(module_object)
                    if file_content:
                        file_hash = hashlib.sha256(file_content).hexdigest()
                        file_hashes[file_hash] = inspect.getsourcefile(module_object)

                if module_filename:
                    module_identifier = self._get_module_identifier(module_object, module_filename)

                    if module_identifier not in modules_blacklist:
                        if self._does_module_match_notification(module_filename, location, self.external_paths):
                            match_found = True
                            modules_blacklist.append(module_identifier)
                            location.module_found(module_object)
                    else:
                        location.send_warning(
                            Error(exc=RookDuplicatedModule(module_key, module_filename)))

        if not match_found:
            if location.file_hash in file_hashes:
                location.send_warning(Error(exc=RookSourceFilePathSuggestion(filepath, file_hashes[location.file_hash])))

    def remove_aug(self, aug_id):
        location = self._post_import_notifications.pop(aug_id, None)
        if location is None:
            return

        location.set_removed()

    def clear_augs(self):
        # This does not require a lock - `notifications` may be added to or removed from in a different thread
        # while we still haven't replaced it with an empty dict(),
        # but `notifications` is just a pointer to the same dict object pointed to by
        # `_post_import_notifications`. Once we do replace `_post_import_notifications`
        # with a new dict, we can be sure that no more notifications will be added.
        # At that point, it's safe to iterate over `notifications` and remove all notifications.
        notifications = self._post_import_notifications
        self._post_import_notifications = dict()
        for location in six.itervalues(notifications):
            location.set_removed()

    def pre_fork(self):
        CountingImportLock().__enter__()

    def post_fork(self):
        try:
            CountingImportLock().__exit__()
        except RuntimeError:
            # This is a bit of a heck as we don't seem to hold import lock in the child.
            # See the TODO about migrating away from the import lock.
            pass

    def _is_valid_module(self, module_object, module_filename):
        return module_filename and os.path.splitext(module_filename)[1] in _SUPPORTED_PY_FORMATS and \
            module_object and inspect.ismodule(module_object) and hasattr(module_object, '__file__')

    def _query_thread(self):
        logger.debug('Starting ImportService thread')
        self._bdb_location_service.ignore_current_thread()

        while not self._quit:
            try:
                with CountingImportLock():
                    self.evaluate_module_list()
            except:  # lgtm[py/catch-base-exception]
                if logger:
                    logger.exception("Error while evaluating module list")

            # time can be None if interpreter is in shutdown
            if not time:
                return
            time.sleep(ImportServiceConfig.SYS_MODULES_QUERY_INTERVAL)

    def evaluate_module_list(self):
        try:
            # No new modules
            if len(self._modules) == len(sys.modules):
                return

            # Get a fresh list
            modules = sys.modules.copy()

            if self._post_import_notifications:
                # self._modules is only replaced (it's immutable - a frozenset), so there's no need to copy it here,
                # just keep a reference to the current self._modules so we don't start working on a different set
                # mid-loop
                old_modules = self._modules

                # For everybody not in the old list, check notifications
                for module_name, module in six.iteritems(modules):
                    module_filename = self._get_module_path(module)
                    if module_name not in old_modules.keys() and self._is_valid_module(module, module_filename):
                        self._notify_of_new_module(module, module_filename)

            # Update the "old" list
            self._modules = modules

        except Exception:
            logger.exception("Exception in ImportService")

    def _notify_of_new_module(self, module_object, module_filename):
        self._trigger_all_notifications_for_module(module_filename, module_object)

    def _trigger_all_notifications_for_module(self, module_filename, module):
        if module_filename:
            for aug_id, location in six.iteritems(self._post_import_notifications.copy()):
                if self._does_module_match_notification(module_filename, location,
                                                        self.external_paths):
                    try:
                        location.module_found(module)
                    except:  # lgtm[py/catch-base-exception]
                        logger.exception("Error on module load callback")
                elif os.path.basename(location.filename) == os.path.basename(module_filename):
                    file_content = self.file_utils.get_safe_file_contents(module)
                    if file_content:
                        file_hash = hashlib.sha256(file_content).hexdigest()
                        if location.file_hash == file_hash:
                            location.send_warning(Error(exc=RookSourceFilePathSuggestion(location.filename, module_filename)))

    def _get_module_path(self, module):
        if module is None or not hasattr(module, '__name__') or not hasattr(module, '__file__'):
            return None
        result = self._path_cache.get(module.__name__)

        if result is not None:
            return result

        if module:
            try:
                path = inspect.getsourcefile(module)
                if not path:
                    if module.__file__.endswith('.pyc'):
                        path = module.__file__.replace('.pyc', '.py')
            except Exception:
                return None

            if path:
                result = os.path.normcase(os.path.abspath(path))
                self._path_cache[module.__name__] = result
                return result

        return None

    @staticmethod
    def _get_module_identifier(module, filename):
        module_identifier = str(filename)
        module_identifier += "#" if not hasattr(module, '__name__') else str(module.__name__)
        module_identifier += "#" if not hasattr(module, '__loader__') else str(type(module.__loader__))
        module_identifier += "#" if not (hasattr(module, '__loader__') and hasattr(module.__loader__, 'path')) \
            else str(module.__loader__.path)

        return module_identifier

    @staticmethod
    def is_black_listed_path(path):
        for blacklisted_path in _BLACKLISTED_PATHS:
            if path.startswith(blacklisted_path):
                return True
        return False

    @staticmethod
    def _does_module_match_notification(module_filename, location, external_paths):
        if not location.include_externals:
            for external_path in external_paths:
                if external_path in module_filename:
                    return False

        if location.filename and ImportService.path_contains_path(module_filename, location.filename) and \
                        not ImportService.is_black_listed_path(module_filename):
            return True
        else:
            return False

    @staticmethod
    def path_contains_path(full_path, partial_path):
        if full_path.endswith(partial_path):
            return len(full_path) == len(partial_path) or full_path[-len(partial_path)-1] in ('/', '\\')
        else:
            return False



import time
import threading

from rook.config import ShutdownConfig

class MonitorService(object):

    class Usage(object):
        def __init__(self, cpu, global_cpu, virtual_memory, start_time, uptime):
            self.cpu = cpu
            self.global_cpu = global_cpu
            self.virtual_memory = virtual_memory
            self.start_time = start_time
            self.uptime = uptime

    NAME = "Monitor"

    def __init__(self):
        self._thread = None
        self._running = False
        self.last_usage = None
        self._psutil = None

    def start(self):
        import psutil
        self._psutil = psutil

        if not self._thread:
            self._thread = threading.Thread(name="rookout-" + type(self).__name__, target=self._collect_data)
            self._thread.daemon = True

            self._running = True
            self._thread.start()

    def stop(self):
        self._running = False
        self._thread.join(1)

    @staticmethod
    def get_last_usage():
        return monitor_instance.last_usage

    @staticmethod
    def get_instance():
        return monitor_instance

    def _collect_data(self):
        process = self._psutil.Process()

        times = process.cpu_times()
        last_user_time = times.user

        cpu_count = self._psutil.cpu_count()

        start_time = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(process.create_time()))
        uptime = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(self._psutil.boot_time()))

        while self._running:
            try:
                if ShutdownConfig.IS_SHUTTING_DOWN:
                    return

                global_cpu = self._psutil.cpu_percent()

                times = process.cpu_times()
                current_user_time = times.user

                diff = current_user_time - last_user_time
                process_cpu = diff / (500 * cpu_count) * 100000

                last_user_time = current_user_time

                mem_info = process.memory_info()
                virtual_memory = mem_info.vms

                self.last_usage = MonitorService.Usage(process_cpu, global_cpu, virtual_memory, start_time, uptime)

            except:  # we want the loop to always keep going lgtm[py/catch-base-exception]
                pass

            time.sleep(0.5)


monitor_instance = MonitorService()



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: variant2.proto

from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2
from . import variant_pb2 as variant__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='variant2.proto',
  package='com.rookout',
  syntax='proto3',
  serialized_options=b'\n\031com.rookout.rook.protobufZ2github.com/Rookout/management/pkg/protos/protobuf2',
  serialized_pb=b'\n\x0evariant2.proto\x12\x0b\x63om.rookout\x1a\x1fgoogle/protobuf/timestamp.proto\x1a\rvariant.proto\"\xa0\x01\n\x06\x45rror2\x12\x0f\n\x07message\x18\x01 \x01(\t\x12\x0c\n\x04type\x18\x02 \x01(\t\x12)\n\nparameters\x18\x03 \x01(\x0b\x32\x15.com.rookout.Variant2\x12\"\n\x03\x65xc\x18\x04 \x01(\x0b\x32\x15.com.rookout.Variant2\x12(\n\ttraceback\x18\x05 \x01(\x0b\x32\x15.com.rookout.Variant2\"\xe1\x04\n\x08Variant2\x12\x1e\n\x16variant_type_max_depth\x18\x01 \x01(\r\x12$\n\x1coriginal_type_index_in_cache\x18\x02 \x01(\r\x12 \n\x18\x61ttribute_names_in_cache\x18\x03 \x03(\r\x12/\n\x10\x61ttribute_values\x18\x04 \x03(\x0b\x32\x15.com.rookout.Variant2\x12\x15\n\roriginal_size\x18\x05 \x01(\r\x12\x12\n\nlong_value\x18\x06 \x01(\x03\x12\x1c\n\x14\x62ytes_index_in_cache\x18\x07 \x01(\r\x12.\n\x0f\x63ollection_keys\x18\x08 \x03(\x0b\x32\x15.com.rookout.Variant2\x12\x30\n\x11\x63ollection_values\x18\t \x03(\x0b\x32\x15.com.rookout.Variant2\x12\x14\n\x0c\x64ouble_value\x18\n \x01(\x01\x12\x34\n\x0b\x63ode_values\x18\x0b \x03(\x0b\x32\x1f.com.rookout.Variant.CodeObject\x12.\n\ntime_value\x18\x0c \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12(\n\x0b\x65rror_value\x18\x10 \x01(\x0b\x32\x13.com.rookout.Error2\x12\x33\n\rcomplex_value\x18\x11 \x01(\x0b\x32\x1c.com.rookout.Variant.Complex\x12\x36\n\x08livetail\x18\x12 \x01(\x0b\x32$.com.rookout.Variant.LiveTailMessageBO\n\x19\x63om.rookout.rook.protobufZ2github.com/Rookout/management/pkg/protos/protobuf2b\x06proto3'
  ,
  dependencies=[google_dot_protobuf_dot_timestamp__pb2.DESCRIPTOR,variant__pb2.DESCRIPTOR,])




_ERROR2 = _descriptor.Descriptor(
  name='Error2',
  full_name='com.rookout.Error2',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='message', full_name='com.rookout.Error2.message', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='type', full_name='com.rookout.Error2.type', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='parameters', full_name='com.rookout.Error2.parameters', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='exc', full_name='com.rookout.Error2.exc', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='traceback', full_name='com.rookout.Error2.traceback', index=4,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=80,
  serialized_end=240,
)


_VARIANT2 = _descriptor.Descriptor(
  name='Variant2',
  full_name='com.rookout.Variant2',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='variant_type_max_depth', full_name='com.rookout.Variant2.variant_type_max_depth', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='original_type_index_in_cache', full_name='com.rookout.Variant2.original_type_index_in_cache', index=1,
      number=2, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='attribute_names_in_cache', full_name='com.rookout.Variant2.attribute_names_in_cache', index=2,
      number=3, type=13, cpp_type=3, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='attribute_values', full_name='com.rookout.Variant2.attribute_values', index=3,
      number=4, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='original_size', full_name='com.rookout.Variant2.original_size', index=4,
      number=5, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='long_value', full_name='com.rookout.Variant2.long_value', index=5,
      number=6, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='bytes_index_in_cache', full_name='com.rookout.Variant2.bytes_index_in_cache', index=6,
      number=7, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='collection_keys', full_name='com.rookout.Variant2.collection_keys', index=7,
      number=8, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='collection_values', full_name='com.rookout.Variant2.collection_values', index=8,
      number=9, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='double_value', full_name='com.rookout.Variant2.double_value', index=9,
      number=10, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='code_values', full_name='com.rookout.Variant2.code_values', index=10,
      number=11, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='time_value', full_name='com.rookout.Variant2.time_value', index=11,
      number=12, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='error_value', full_name='com.rookout.Variant2.error_value', index=12,
      number=16, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='complex_value', full_name='com.rookout.Variant2.complex_value', index=13,
      number=17, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='livetail', full_name='com.rookout.Variant2.livetail', index=14,
      number=18, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=243,
  serialized_end=852,
)

_ERROR2.fields_by_name['parameters'].message_type = _VARIANT2
_ERROR2.fields_by_name['exc'].message_type = _VARIANT2
_ERROR2.fields_by_name['traceback'].message_type = _VARIANT2
_VARIANT2.fields_by_name['attribute_values'].message_type = _VARIANT2
_VARIANT2.fields_by_name['collection_keys'].message_type = _VARIANT2
_VARIANT2.fields_by_name['collection_values'].message_type = _VARIANT2
_VARIANT2.fields_by_name['code_values'].message_type = variant__pb2._VARIANT_CODEOBJECT
_VARIANT2.fields_by_name['time_value'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_VARIANT2.fields_by_name['error_value'].message_type = _ERROR2
_VARIANT2.fields_by_name['complex_value'].message_type = variant__pb2._VARIANT_COMPLEX
_VARIANT2.fields_by_name['livetail'].message_type = variant__pb2._VARIANT_LIVETAILMESSAGE
DESCRIPTOR.message_types_by_name['Error2'] = _ERROR2
DESCRIPTOR.message_types_by_name['Variant2'] = _VARIANT2
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

Error2 = _reflection.GeneratedProtocolMessageType('Error2', (_message.Message,), {
  'DESCRIPTOR' : _ERROR2,
  '__module__' : 'variant2_pb2'
  # @@protoc_insertion_point(class_scope:com.rookout.Error2)
  })
_sym_db.RegisterMessage(Error2)

Variant2 = _reflection.GeneratedProtocolMessageType('Variant2', (_message.Message,), {
  'DESCRIPTOR' : _VARIANT2,
  '__module__' : 'variant2_pb2'
  # @@protoc_insertion_point(class_scope:com.rookout.Variant2)
  })
_sym_db.RegisterMessage(Variant2)


DESCRIPTOR._options = None
# @@protoc_insertion_point(module_scope)



import datetime

from .dumped_object_namespace import DumpedObjectNamespace


# TODO - remove this class
class DumpedPrimitiveNamespace(DumpedObjectNamespace):

    def __init__(self, obj, type, common_type, attributes={}, methods=()):
        super(DumpedPrimitiveNamespace, self).__init__(type, common_type, attributes, methods)
        self.obj = obj

    def _get_value_for_json(self):
        if isinstance(self.obj, complex):
            return {u'real': self.obj.real, u'imag': self.obj.imag}
        elif isinstance(self.obj, datetime.datetime):
            return str(self.obj)
        else:
            return self.obj



"""
from https://github.com/quick2wire/quick2wire-python-api/blob/master/quick2wire/syscall.py
"""


import errno
import ctypes

libc = ctypes.CDLL(None, use_errno=True)


def errcheck(result, func, args):
    if result < 0:
        e = ctypes.get_errno()
        raise OSError(e, errno.strerror(e))
    return result


def lookup(restype, name, argtypes):
    f = libc[name]
    f.restye = restype
    f.argtypes = argtypes
    f.errcheck = errcheck
    return f


class SelfClosing(object):
    def __enter__(self):
        return self

    def __exit__(self, *exc):
        self.close()
        return False



from .python_object_namespace import PythonObjectNamespace
from .dumped_primitive_namespace import DumpedPrimitiveNamespace


class StringNamespace(DumpedPrimitiveNamespace):

    def __init__(self, obj, original_size, type, common_type, attributes={}):
        super(StringNamespace, self).__init__(obj, type, common_type, attributes, self.METHODS)
        self.original_size = original_size

    def size(self, args):
        return PythonObjectNamespace(len(self.obj))

    def original_size(self, args):
        return PythonObjectNamespace(self.original_size)

    METHODS = (size, original_size)



from .location import Location

from rook.services.exception_capturing_location_service import ExceptionCapturingLocationService


class LocationExceptionHandler(Location):

    NAME = 'exception_handler'

    def __init__(self, output, aug, arguments, processor_factory):
        super(LocationExceptionHandler, self).__init__(output, aug)

    def _add_aug_internal(self, trigger_services):
        trigger_services.get_service(ExceptionCapturingLocationService.NAME).add_exception_capturing_aug(self)



import datetime
import json as py_json


class JSONEncoder(py_json.JSONEncoder):

    def __init__(self, fail_on_unknown=False, **kwargs):
        kwargs['skipkeys'] = True
        super(JSONEncoder, self).__init__(**kwargs)

        self.fail_on_unknown = fail_on_unknown

    def default(self, obj):
        if isinstance(obj, datetime.datetime):
            return obj.isoformat()

        if self.fail_on_unknown:
            return super(JSONEncoder, self).default(obj)
        else:
            return None


class json(object):

    @staticmethod
    def loads(s, **kwargs):
        return py_json.loads(s, **kwargs)

    @staticmethod
    def dumps(obj, **kwargs):
        return py_json.dumps(obj, cls=JSONEncoder, **kwargs)

    @staticmethod
    def dump(obj, fp, **kwargs):
        return py_json.dump(obj, fp, cls=JSONEncoder, **kwargs)




import six
import sys
import threading
import os

from .namespace import Namespace

from .python_object_namespace import PythonObjectNamespace
from .container_namespace import ContainerNamespace
from .collection_namespace import ListNamespace
from .frame_namespace import FrameNamespace

from .stack_namespace import StackNamespace

from ...logger import logger


class PythonUtilsNamespace(Namespace):

    def __init__(self):
        super(PythonUtilsNamespace, self).__init__(self.METHODS)

    def read_attribute(self, name):
        raise NotImplementedError()

    def exception(self, args):
        return PythonObjectNamespace(sys.exc_info()[1])

    def exception_string(self, args):
        if six.PY2:
            return PythonObjectNamespace(unicode(sys.exc_info()[1]))
        else:
            return PythonObjectNamespace(str(sys.exc_info()[1]))

    def exception_type(self, args):
        return PythonObjectNamespace(sys.exc_info()[0])

    def module(self, args):
        return PythonObjectNamespace(sys.modules[args])

    def thread_id(self, args):
        return PythonObjectNamespace(threading.currentThread().ident)

    def thread_name(self, args):
        return PythonObjectNamespace(threading.currentThread().name)

    def threads(self, args):
        return PythonObjectNamespace(threading.enumerate())

    def thread_tracebacks(self, args):
        threads = threading.enumerate()
        frames = sys._current_frames()

        result = []
        for thread in threads:
            try:
                traceback = StackNamespace(FrameNamespace(frames[thread.ident])).traceback(None)
            except KeyError:
                traceback = PythonObjectNamespace(None)

            result.append(ContainerNamespace({
                'id': PythonObjectNamespace(thread.ident),
                'name': PythonObjectNamespace(thread.name),
                'daemon': PythonObjectNamespace(thread.daemon),
                'traceback': traceback,
            }))

        return ListNamespace(result)

    def env(self, args):
        try:
            value = os.environ[args]
        except KeyError:
            logger.exception("Failed to get env variable")
            value = None

        return PythonObjectNamespace(value)

    METHODS = (exception, exception_string, exception_type, module, thread_id, thread_name, threads, thread_tracebacks, env)



import os
import sys
import traceback
from rook.processor.namespaces.container_namespace import ContainerNamespace
from rook.processor.namespaces.python_object_namespace import PythonObjectNamespace
from rook.processor.namespaces.collection_namespace import ListNamespace
from rook.services.exception_capturing_location_service import ExceptionCapturingLocationService
from rook.logger import logger
import uuid
from rook.singleton import singleton_obj

if sys.version.startswith('2'):
    import funcsigs
    signature_function = funcsigs.signature
else:
    import inspect
    signature_function = inspect.signature


class ExceptionCollector(object):

    class Traceback(object):
        def __init__(self, filename, line_number, function_name):
            self.filename = filename
            self.line = line_number
            self.function = function_name

    def __init__(self, output=None, exception_capturing_service=None):
        if output is None:
            output = singleton_obj._output

        self._output = output
        if exception_capturing_service is None:
            exception_capturing_service = singleton_obj.get_trigger_services().get_service(
                    ExceptionCapturingLocationService.NAME)

        self._exception_capturing_service = exception_capturing_service

    def collect(self, exc=None):
        try:
            if not self._exception_capturing_service.activation_status():
                return

            if exc is None:
                exc = sys.exc_info()[1]

            data_namespace = self._get_exception_data(exc)
            aug_ids = self._get_global_rule_ids()
            message_id = self._get_message_id()

            for aug_id in aug_ids:
                self._output.send_user_message(aug_id, message_id, data_namespace)
        except Exception as exc:
            logger.exception('Rook capture exception failed while collecting data')


    def _get_exception_data(self, exc):
        trace = self._get_exception_traceback(exc)
        local_variables = sys._getframe(4).f_locals
        return self._get_rookout_container_namespace(local_variables, trace, exc)

    def _get_exception_traceback(self, exc):
        if hasattr(exc, '__traceback__'):
            return self._get_exception_traceback_py3(exc)
        return self._get_exception_traceback_py2()

    def _get_exception_traceback_py3(self, exc):
        new_traceback = []
        trace = traceback.extract_tb(exc.__traceback__)

        for frame in trace:
            new_traceback.append(self.Traceback(frame.filename, frame.lineno, frame.name))

        return new_traceback

    def _get_exception_traceback_py2(self):
        new_traceback = []
        raw_traceback = sys.exc_info()[2]
        trace = traceback.extract_tb(raw_traceback)

        for frame in trace:
            new_traceback.append(self.Traceback(frame[0], frame[1], frame[2]))

        return new_traceback

    def _get_rookout_container_namespace(self, local_variables, traceback, exc):

        frame_container = self._get_frame_container(local_variables)
        traceback_container = self._wrap_traceback_in_namespace(traceback)
        exception_information_namespace = self._get_exception_information_namespace(exc, traceback)
        rookout_container = ContainerNamespace(
            {'frame': frame_container, 'traceback': traceback_container,
             'exception_information': exception_information_namespace})

        return ContainerNamespace({'rookout': rookout_container})

    def _get_frame_container(self, local_variables):
        locals_container = self._wrap_locals_in_container_namespace(local_variables)
        return ContainerNamespace({'locals': locals_container})

    def _wrap_locals_in_container_namespace(self, local_variables):
        """
        returns a ContainerNamespace with each variable given wrapped in PythonObjectNamespace
        """
        locals_container = ContainerNamespace({})

        for local_variable_name in local_variables:
            local_container_variable = PythonObjectNamespace(local_variables[local_variable_name])
            locals_container.write_attribute(local_variable_name, local_container_variable)

        return locals_container

    def _wrap_traceback_in_namespace(self, traceback_stack):
        traceback_attributes = []

        for stack_cell in traceback_stack:
            stack_cell_container = ContainerNamespace({})

            for attribute_name in [attr_name for attr_name in dir(stack_cell) if not attr_name.startswith('_')]:
                attribute_container = PythonObjectNamespace(stack_cell.__getattribute__(attribute_name))
                stack_cell_container.write_attribute(attribute_name, attribute_container)

            traceback_attributes.append(stack_cell_container)

        return ListNamespace(traceback_attributes)

    def _get_exception_information_namespace(self, exc, traceback_stack):
        exception_type = PythonObjectNamespace(type(exc).__name__)
        exception_frame = traceback_stack[-1]
        file_path = PythonObjectNamespace(exception_frame.filename)
        filename = PythonObjectNamespace(os.path.split(exception_frame.filename)[1])
        line_number = PythonObjectNamespace(exception_frame.line)
        try:
            exception_parameters = PythonObjectNamespace(dict(zip(
                [parameter for parameter in signature_function(exc.__init__).parameters],
                exc.args)))
        except ValueError:
            # init signature is not supported for this type of exception in Python2
            exception_parameters = PythonObjectNamespace({})

        return ContainerNamespace({
            'exception_type': exception_type,
            'file_path': file_path,
            'filename': filename,
            'line_number': line_number,
            'exception_parameters': exception_parameters})

    def _get_global_rule_ids(self):
        return self._exception_capturing_service.get_ids()

    def _get_message_id(self):
        return uuid.uuid4().hex


def collect(exc):
    ExceptionCollector().collect(exc)



import json
from rook.config import update_config

class CommandHandler(object):
    def __init__(self, agent_com, aug_manager):
        self._agent_com = agent_com
        self._aug_manager = aug_manager

        self._agent_com.on("InitialAugsCommand", self._handle_initial_augs)
        self._agent_com.on("AddAugCommand", self._handle_add_aug)
        self._agent_com.on("RemoveAugCommand", self._handle_remove_aug)

    def _handle_initial_augs(self, initial_augs):
        update_config(initial_augs.sdk_configuration)

        augs = [json.loads(aug_json) for aug_json in initial_augs.augs]
        self._aug_manager.initialize_augs(augs)

    def _handle_add_aug(self, command):
        self._aug_manager.add_aug(json.loads(command.aug_json))

    def _handle_remove_aug(self, command):
        self._aug_manager.remove_aug(command.aug_id)

    def __del__(self):
        pass



import threading

from six.moves.queue import Queue, Empty

from rook.com_ws.eventfd.eventfd import Semaphore as SemaphoreEventFD


MAX_QUEUE_SIZE = 2048


class SelectableQueue(object):
    """
    Uses a eventfd as a semaphore for reading and writing to the internal queue.
    This way, this queue can be used as a parameter to select() or poll().
    Maximum queue size is MAX_QUEUE_SIZE.
    """
    def __init__(self):
        self._lock = threading.Lock()
        self._queue = Queue(maxsize=MAX_QUEUE_SIZE)
        self._event_fd = SemaphoreEventFD(blocking=False)

    def __del__(self):
        self._event_fd.close()

    def put(self, item):
        with self._lock:
            if self._queue.full():
                return

            self._event_fd.signal()

            self._queue.put_nowait(item)

    def get(self):
        if not self._event_fd.wait():
            raise Empty()

        with self._lock:
            return self._queue.get_nowait()

    def fileno(self):
        return self._event_fd.fileno()

    def qsize(self):
        return self._queue.qsize()




from .operations_factory import OperationsFactory
from .paths_factory import PathsFactory
from .processor import Processor


class ProcessorFactory(object):

    def __init__(self, extended_operations_factories, extended_path_factories):
        self.operations_factory = OperationsFactory(extended_operations_factories)
        self.paths_factory = PathsFactory(extended_path_factories)

    def register_operation(self, operation):
        self.operations_factory.register_method(operation)

    def get_operation(self, configuration):
        return self.operations_factory.get_operation(configuration, self)

    def get_path(self, configuration):
        return self.paths_factory.get_path(configuration, self)

    def get_processor(self, configuration):
        return Processor(configuration, self)



from .location import Location

from rook.services.http_server_services.http_server_services import HttpServerService


class HttpServerHandler(Location):

    NAME = 'http_server'

    def __init__(self, output, aug, arguments, processor_factory):
        super(HttpServerHandler, self).__init__(output, aug)

    def _add_aug_internal(self, trigger_services):
        trigger_services.get_service(HttpServerService.NAME).add_logging_aug(self)



#!/usr/bin/env python
import platform
import sys
from glob import glob
from os.path import join, dirname
import imp

try:
    # Use setuptools if available, for install_requires (among other things).
    from setuptools import setup, find_packages, Extension  # lgtm[py/import-of-mutable-attribute]
    setuptools = True
except ImportError:
    from distutils.core import setup, find_packages, Extension  # lgtm[py/import-of-mutable-attribute]
    setuptools = False

# Classic setup.py

kwargs = {}

# Version configuration
version = imp.load_source('version', join(dirname(__file__), 'rook', 'version.py')).VERSION

# Readme
with open('README.rst') as f:
    kwargs['long_description'] = f.read()

if sys.platform in ('darwin', 'linux2', 'linux'):
    ext_modules = []
    extra_compile_args = [
        '-std=c++0x',
        '-g0',
        '-O3',
        '-Wno-deprecated-register']
    extra_link_args = []

    if sys.platform == 'darwin':
        mac_version = platform.mac_ver()[0]
        extra_compile_args.append("-mmacosx-version-min=" + mac_version)
        extra_link_args.append("-Xlinker -macosx_version_min " + mac_version)

    if ('CPython' == platform.python_implementation()) and \
            ((sys.version_info[0] == 2 and sys.version_info > (2, 7, 0)) or
             (sys.version_info[0] == 3 and sys.version_info >= (3, 5, 0))):
        ext_modules.append(Extension(
            'rook.services.cdbg_native',
            sources=glob('rook/services/exts/cloud_debug_python/*.cc'),
            extra_compile_args=extra_compile_args, extra_link_args=extra_link_args))

    ext_modules.append(Extension(
        'native_extensions',
        sources=glob('rook/native_extensions/*.cc'),
        extra_compile_args=extra_compile_args, extra_link_args=extra_link_args))

    kwargs['ext_modules'] = ext_modules


if setuptools:
    # If setuptools is not available, you're on your own for dependencies.
    install_requires = [
        "six >= 1.11",
        "websocket-client >= 0.56, != 0.58, != 0.59, != 1.0, != 1.1",
        "psutil >= 5.8.0",
        "certifi",
        "funcsigs"
    ]

    kwargs['install_requires'] = install_requires

setup(
    name="rook",
    version=version,
    packages=find_packages(where='.', exclude=['contrib', 'docs', '*test*']),
    include_package_data=True,
    author="Rookout",
    author_email="liran@rookout.com",
    url="http://rookout.com/",
    description="Rook is a Python package for on the fly debugging and data extraction for application in production",
    license="https://get.rookout.com/SDK_LICENSE.pdf",
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',

        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',

        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',

        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',

    ],
    zip_safe=False,
    extras_require={
        'ssl_backport': ['backports.ssl', 'backports.ssl_match_hostname', 'PyOpenSSL'],
        ':python_version <= "3.5"': ["protobuf >= 3.7.1, < 3.18.0"],
        ':python_version > "3.5"': ["protobuf >= 3.7.1, <= 4.0.0"]
    },
    **kwargs
)
